#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#include "config.h"

#define HASHSIZE 8192
#define TOPSIZE 20

struct node
{
   char *filename;
   time_t last_used;
   unsigned long size;
   struct node *prev,*next,*nexthash;
} *firstnode,*lastnode,
  *hashnode[HASHSIZE],
  *toplist[TOPSIZE];

#ifndef HAVE_STRDUP
char *strdup(const char *s)
{
  char *b;
  b=malloc(strlen(s) + 1);
  strcpy(b,s);
  return s;
}
#endif



unsigned long size=(unsigned long)(~0),garbsize=0;
unsigned long total_cache_size=0;
unsigned long bytes_per_second=1024;

int do_gc;

void help(void)
{
  /* Multiple printfs to workaround bug in some compilers. */

   printf("proxygarb: size <bytes>       - set maximum cache size (garbage limit) (%lu bytes)\n",size);
   printf("garbsize <bytes>   - set minimum garb level (%lu bytes)\n",garbsize);
   printf("bps <bps>          - set bytes per second\n");
   printf("mark <filename>    - mark usage\n");
   printf("add <filename>     - add to database, last usage=last read\n");
   printf("update <filename>  - mark usage and update information\n");
   printf("remove <filename>  - remove <filename> from database\n");
   printf("garb               - do a garbage collect\n");
   printf("info               - show info\n");
   printf("highscore          - list files to be removed next\n");
   printf("encoded_info       - print info, encoded to be read by spider\n");
   printf("quit               - quit this program\n");
   printf("help               - show this text\n");
}

#define bonus(sz) ((sz)/bytes_per_second)
#define score(lu,size) ((lu)-(bonus(size)))

#ifndef HAVE_MEMMOVE
char *MEMMOVE(char *b,const char *a,int s)
{
  char *t;
  if(a>b) for(t=b;s;s--) *(t++)=*(a++);
  if(a<b) for(t=b+s,a+=s;s;s--) *(--t)=*(--a);
  return b;
}
#define memmove MEMMOVE
#endif

void calculate_high(void)
{
   struct node *n;
   unsigned long max,q,i;

   max=(unsigned long)~0;
   for (i=0; i<TOPSIZE; i++) toplist[i]=NULL;

   n=firstnode;
   while (n)
   {
      if ((q=score(n->last_used,n->size))<max)
      {
	 for (i=0; i<TOPSIZE; i++)
	    if (toplist[i]==NULL) 
	       { toplist[i]=n; break; }
	    else if (score(toplist[i]->last_used,
			   toplist[i]->size)>q)
	    {
	       memmove((char *)(toplist+i+1),(char *)(toplist+i),
		       sizeof(struct node*)*(TOPSIZE-i-1));
	       toplist[i]=n;
	       break;
	    }
	 if (toplist[TOPSIZE-1])
	 {
  	    max=score(toplist[TOPSIZE-1]->last_used,toplist[TOPSIZE-1]->size);
	 }
      }
         
      n=n->next;
   }
}

void mark_usage(struct node *n)
{
   time(&(n->last_used));
}

void read_usage(struct node *n)
{
   struct stat s;
   lstat(n->filename,&s);
   n->last_used=s.st_atime;
}

void update_node(struct node *n)
{
   struct stat s;
   total_cache_size-=n->size;
   if (lstat(n->filename,&s))
   {
     fprintf(stderr, 
	     "proxygarb: warning; can't stat %s. Assumed size=0 bytes.\n",
	     n->filename);
      n->size=0;
      return;
   }
   n->size=s.st_size;
   total_cache_size+=n->size;
   do_gc=1;
}

unsigned hash_value(char *s)
{
   unsigned i=0;
   while (*s) i=((i>>23)|(i<<7))+*s,s++;
   return i%HASHSIZE;
}

struct node *find_node(char *filename,struct node ***prevhash)
{
   struct node *n,**pn;
   pn=&(hashnode[hash_value(filename)]);
   while ((n=*pn))
   {
      if (!strcmp(filename,n->filename)) 
          { if (prevhash) *prevhash=pn; return n; }
      pn=&(n->nexthash);
   }
   return NULL;
}

struct node *create_node(char *filename)
{
   struct node *n;
   n=(struct node*)malloc(sizeof(struct node));

   n->filename=(char*)strdup(filename);
   n->size=0;

   n->prev=lastnode;
   n->next=NULL;
   n->nexthash=hashnode[hash_value(filename)];
   hashnode[hash_value(n->filename)]=n;
   if (lastnode) lastnode->next=n; else firstnode=n;
   lastnode=n;

   update_node(n);
   return n;
}

struct node *find_or_create_node(char *filename)
{
   struct node *n;
   n=find_node(filename,NULL);
   if (n) return n;
   return create_node(filename);
}

void remove_file(char *file)
{
   struct node *n;
   struct node **prevhash;

   if (!file) { printf("proxygarb: missing filename\n"); return; }

   n=find_node(file,&prevhash);
   if (!n) return; 
   total_cache_size-=n->size;

   if (n->prev) n->prev->next=n->next; else firstnode=n->next;
   if (n->next) n->next->prev=n->prev; else lastnode=n->prev;
   *prevhash=n->nexthash;
   free(n->filename);
   free(n);
}

void use(char *file)
{
   struct node *n;
   struct stat s;
   if (!file) { printf("proxygarb: missing filename\n"); return; }
   if(strlen(file) && file[0] != '#')
   {
     if (lstat(file,&s)) 
     {
       printf("proxygarb: can't stat \"%s\", skipping\n",file);
       perror("lstat");
       remove_file(file);
       return;
     }
   } else {
     return;
   }
   n=find_or_create_node(file);
   mark_usage(n);
}

void add(char *file)
{
   struct node *n;
   struct stat s;
   if (!file) { printf("proxygarb: missing filename\n"); return; }

   if(strlen(file) && file[0] != '#')
   {
     if (lstat(file,&s)) 
     {
       printf("proxygarb: can't stat %s, skipping\n",file);
       remove_file(file);
       return;
     }
   } else {
     return;
   }
   n=find_or_create_node(file);
   read_usage(n);
}

void update(char *file)
{
   struct node *n;
   if (!file) { printf("proxygarb: missing filename\n"); return; }

   n=find_node(file,NULL);
   if (n) update_node(n);
   else n=create_node(file);
   mark_usage(n);
}

char *timestring(char *buf,time_t t)
{
   if (t>3600*24) sprintf(buf,"%lud",t/(3600*24));
   else if (t>3599) sprintf(buf,"%luh",t/3600);
   else if (t>59) sprintf(buf,"%lum",t/60);
   else sprintf(buf,"%lus",t);
   return buf;
}

void highscore(void)
{
   int i;
   time_t now;
   char buf[10];
   calculate_high();

   time(&now);
   printf("proxygarb:     %-30s%7s %6s %s\n","file","size","age","score");
   for (i=0; i<TOPSIZE && toplist[i]; i++)
   {
      timestring(buf,now-toplist[i]->last_used);
      printf("proxygarb: %2d. %-30s%7lu %6s %lu\n",i,toplist[i]->filename,toplist[i]->size,
	     buf,score(toplist[i]->last_used,toplist[i]->size));
   }
}

void info(void)
{
   unsigned long i;
   struct node *n;
   i=0; n=firstnode; while (n) { n=n->next; i++; };
   printf("proxygarb: in cache.............%lu bytes in %lu files\n",
	  total_cache_size, i);
   printf("proxygarb: max bytes in cache...%lu bytes\n",size);
   printf("proxygarb: minimum garbage......%lu bytes\n",garbsize);
}


void encoded_info(char *file)
{
   unsigned long i;
   struct node *n;
   time_t now;
   FILE *fp;

   if (!file) { printf("proxygarb: missing filename\n"); return; }

   fp=fopen(file,"wt");
   if (!fp) return;

   calculate_high();

   i=0; n=firstnode; while (n) { n=n->next; i++; };
   fprintf(fp,"%lu\n%lu\n%lu\n%lu\n%lu\n",total_cache_size,i,size,garbsize,bytes_per_second);

   time(&now);
   for (i=0; i<TOPSIZE && toplist[i]; i++)
   {
      fprintf(fp,"%s///%lu///%lu///%lu\n",toplist[i]->filename,toplist[i]->size,
	      toplist[i]->last_used,now-score(toplist[i]->last_used,toplist[i]->size));
   }
   fprintf(fp,"end<tm>");
   fclose(fp);
}

void list(void)
{
   struct node *n;
   time_t now;
   char buf[10];

   time(&now);
   n=firstnode;
   printf("proxygarb: %-30s%7s%5s %s\n","file","size","age","score");
   while (n)
   {
      printf("proxygarb: %-30s%7lu %6s %lu\n",n->filename,n->size,
	     timestring(buf,now-n->last_used),
	     score(n->last_used,n->size));
      n=n->next;
   }
}

void garb_remove(struct node *n)
{
#ifdef HAVE_REMOVE
   remove(n->filename);
#else
   unlink(n->filename);
#endif
   remove_file(n->filename);
}

void garb(void)
{
   int i;
   unsigned long garbed=0;
   if (total_cache_size<size) return;
   for (;total_cache_size && total_cache_size>size-garbsize;)
   {
      calculate_high();
      for (i=0;total_cache_size>size-garbsize && i<TOPSIZE; i++)
      {
	 garbed+=toplist[i]->size;
         garb_remove(toplist[i]);
      }
   }
}

void set_size(char *parm)
{
   if (parm)
     size=atoi(parm);
   garb();
}

void set_garbsize(char *parm)
{
   if (parm) garbsize=atol(parm);
}

void set_bps(char *parm)
{
   if (parm) bytes_per_second=atol(parm);
}

int command(FILE *fp);

void command_file(char *file)
{
   FILE *fp;
   fp=fopen(file,"rt");
   if (!fp) printf("proxygarb: can't read command file %s\n",file);
   else
   {
      while (command(fp));
      fclose(fp);
   }
}

int command(FILE *fp)
{
   char buf[2048];
   char *parm;
   int len;

   if (!fgets(buf,2048,fp)) return 0;
   for (parm=buf; *parm && *parm!='\n'; parm++); *parm=0;
   for (len=0; buf[len]&&buf[len]!=' '; len++);
   if (buf[len]) parm=buf+len+1; else parm=NULL;

#define iscmd(S) (len==strlen(S) && strncmp(buf,(S),len)==0)

   do_gc=0;

   if (iscmd("update")) update(parm);
   else if (iscmd("mark")) use(parm);
   else if (iscmd("add")) add(parm);
   else if (iscmd("remove")) remove_file(parm);
   else if (iscmd("info")) info();
   else if (iscmd("encoded_info")) encoded_info(parm);
   else if (iscmd("list")) list();
   else if (iscmd("size")) set_size(parm);
   else if (iscmd("bps")) set_bps(parm);
   else if (iscmd("garbsize")) set_garbsize(parm); /* Per */
   else if (iscmd("bytes_per_second")) set_bps(parm);
   else if (iscmd("garb")) garb();
   else if (iscmd("highscore")) highscore();
   else if (iscmd("help")) help();
   else if (iscmd("command")) command_file(parm);
   else if (iscmd("quit")) return 0;
   else add(buf);

   if (do_gc) garb();

   return 1;
}

int main(int ac,char **am)
{
   int i;
   FILE *fp;

   fprintf(stderr,"proxygarb 1.0 (c) Informationsvvarna 1995\n");

   for (i=0; i<HASHSIZE; i++) 
     hashnode[i]=NULL;
   firstnode=lastnode=NULL;

   for (i=1; i<ac; i++)
   {
      fp=fopen(am[i],"rt");
      if (!fp) printf("proxygarb: can't read command file %s\n",am[i]);
      else
      {
	 while (command(fp));
	 fclose(fp);
      }
   }

   while (command(stdin));
   return 0;
}
