#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef HAVE_DIRENT_H
#include <dirent.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "config.h"

#ifndef TRUE
#define TRUE -1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define LINK      1
#undef FILE
#define FILE      2
#define DIRECTORY 3 
#define SPECIAL   4

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif



#define MAXIMALPATHLENGTH MAXPATHLEN

char *existinode(char* directory,char* file)
{
  int result;
  struct stat buf;
  char* inode;
  char* nothing;

  inode=(char*)malloc((strlen(directory)+strlen(file)+2)*sizeof(char));
  nothing=strcpy(inode,directory);
  if (directory[0]!=(char)0)
    strcat(inode,"/");
  strcat(inode,file);
  result=stat(inode,&buf);
  if (result==0)
    return(inode);
  free(inode); 
  return(NULL);
}  

/*
 * identify inode, if it is link, directory, file or special 
 */

int identifyinode(char* inode)
{
  int result;
  struct stat buf;
  /* 
     note, that it's important to use here lstat instead of stat under POSIX,
     cause stat can't get back the st_mode of the link itself (it gets the
     st_mode of the pointed inode)
   */
  result=lstat(inode,&buf);
  if (result==-1)
    return(0);   
  if      (((buf.st_mode) & S_IFLNK)==S_IFLNK) result=LINK;
  else if (((buf.st_mode) & S_IFDIR)==S_IFDIR) result=DIRECTORY;
  else if (((buf.st_mode) & S_IFREG)==S_IFREG) result=FILE;
  else                                         result=SPECIAL;
  return(result);
}

/*
 * test if the file is executable, if yes, return inode, if no return NULL
 */

char* testexecutable(char* inode)
{
  int result;
  struct stat buf;

  result=stat(inode,&buf);
  if (result<0)
    return(NULL);      
  result=-1;
  if      (((buf.st_mode) & S_IXUSR) != 0) result=0;
  else if (((buf.st_mode) & S_IXGRP) != 0) result=0;
  else if (((buf.st_mode) & S_IXOTH) != 0) result=0;
  if (result<0)
    return(NULL);   
  return(inode);
}

/*
 * search the inode of a filelink
 */

char* inodeoflink(char* inode)
{
  int result;
  int i;
  size_t buflen=MAXIMALPATHLENGTH;
  char* buf;

  buf=(char*)malloc((MAXIMALPATHLENGTH+1)*sizeof(char));
  for (i=0;i<=MAXIMALPATHLENGTH;i=i+1)
    buf[i]=(char)0;
  result=readlink(inode,buf,buflen);
  buf[buflen]=(char)0;
  if (result<0)
    return(NULL);   
  return(buf);
}

char *getdir(char* file)
{
  int i;
  if (file==NULL) 
    return(file);
  for (i=strlen(file)-1;i>=0;i=i-1)
    if (file[i]=='/')
    { 
      file[i+1]=(char)0; 
      break;
    }
  if (i<0)
    return(NULL);
  return(file);
}

/* 
 * get a proper path of the inode of a FILElink
 */

char* getlink(char* inode)
{
  char* inodepath;
  char* temp;
  char* returninode;

  returninode=inodeoflink(inode);
  if (returninode==NULL)
  {
    return(NULL);
    free(inode);
  }   
  /* relative links need extra code ! */
  if (returninode[0]!='/')
  {
    inodepath=getdir(inode); 
    temp=malloc(strlen(inodepath)+strlen(returninode)+1);
    strcpy(temp,inodepath); 
    strcat(temp,returninode);
    returninode=temp;
  }
  free(inode);
  return(returninode);
}

/* 
 * look if the inode pointed by directory and file is a true file 
 * if the inode is a filelink, search the true file
 * if the inode is a directory or a special file, return error (NULL)
 * on success (file found) and it's executable return the path
 */

char* searchfile(char* inode)
{
  while (identifyinode(inode)==LINK)
    inode=getlink(inode);
  if (inode!=NULL) 
    if (identifyinode(inode)!=FILE) 
    {
      free(inode);
      return(NULL);
    }
  inode=testexecutable(inode);
  return(inode);
}

static char *combine_path(char *cwd,char *file)
{
  /* cwd is supposed to be combined already */
  char *ret;
  register char *from,*to;
  char *my_cwd;
  my_cwd=0;
  

  if(file[0]=='/')
  {
    cwd="/";
    file++;
  }else{
    
    if(!cwd)
    {
#if HAVE_GETWD
      cwd=(char *)getwd(my_cwd=(char *)malloc(MAXPATHLEN+1));
      if(!cwd)
      {
	fprintf(stderr, "Couldn't fetch current path.\n");
	abort();
      }
#else
#if HAVE_GETCWD
      my_cwd=cwd=(char *)getcwd(0,1000); 
#else
      /* maybe autoconf was wrong.... if not, insert your method here */
      my_cwd=cwd=(char *)getcwd(0,1000); 
#endif
#endif
    }
  }

  if(cwd[strlen(cwd)-1]=='/')
  {
    ret=(char *)malloc(strlen(cwd)+strlen(file)+1);
    strcpy(ret,cwd);
    strcat(ret,file);
  }else{
    ret=(char *)malloc(strlen(cwd)+strlen(file)+2);
    strcpy(ret,cwd);
    strcat(ret,"/");
    strcat(ret,file);
  }

  from=to=ret;
  
  while(( *to = *from ))
  {
    if(*from == '/')
    {
      while(from[1] == '/') from++;
      if(from[1] == '.')
      {
	switch(from[2])
	{
	 case '.':
	  if(from[3] == '/' || from[3] == 0)
	  {
	    from+=3;
	    if(to != ret)
	    {
	      for(--to;*to!='/' && to>ret;to--);
	    }
	    if(!*from && to==ret && *to=='/') to++;
	  }
	  continue;

	 case '/':
	 case 0:
	  from+=2;
	  continue;
	}
      }
    }
    from++;
    to++;
  }

  if(my_cwd) free(my_cwd);
  return ret;
}

/* 
 * detects the path of the called binary from argv[0] and the enviroment-
 * variable $PATH (the path includes the last slash of the file) 
 */

char *selfdir(char *argv0)
{
  char *target, *nothing, *dollarpath, *temp, *inode;
  int pathlen, i;

  if(argv0[0] != '.')
    for (i=0;i<=strlen(argv0);i++)
    {
      if (argv0[i]=='/') 
      {
	target=(char*)malloc((strlen(argv0)+1)*sizeof(char));
	strcpy(target, argv0);
	return(getdir(searchfile(target)));
      }
    }
  else
    return getdir(combine_path(0, argv0));

  temp=getenv("PATH");
  pathlen=strlen(temp);
  dollarpath=(char*)malloc((pathlen+1)*sizeof(char));
  nothing=strcpy(dollarpath,temp);

  for(i=0;i<pathlen;i++) 
    if((*(dollarpath+i)==':') || (*(dollarpath+i)<=' '))
      *(dollarpath+i)=(char)0;

  i=0;
  while (i<pathlen)
    if(*(dollarpath+i)!=(char)0)
    {
      inode=existinode(dollarpath+i,argv0);  
      if(inode!=NULL)  
      {          
	target=getdir(searchfile(inode));
	if(target!=NULL)
	{
	  free(dollarpath);
	  return(target);
	}
      }
      i=i+strlen(dollarpath+i); 
    }
    else
      i=i+1;
  free(dollarpath);
  return(NULL);
}

void main(int wc, char **argv)
{
  char *c;
  c=selfdir(argv[0]);
  if(c)
    puts(c);
}
