/* Shuffle data from A to B. 
 * -------------------------
 * SOLARIS ONLY 
 *
 * This file use solaris threads and mmapped io for maximum 
 * performance. It could be used as an freaked up 'cp' command..
 *
 * It can only be used with Solaris 2.4 and up. 
 * And it can probably only be compiled with gcc.. :-)
 *
 * (Yes, I know, not exactly portable.) 
 *
 * This file is part of the Spinner WWW server, but is not in any 
 * way nessesary to run it. To use this one, do 'make shuffle-install'.
 */

#define _REENTRANT

#include <stdio.h>
#include <stdlib.h>
#include <thread.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/uio.h>

#include <sys/sockio.h>
#include <sys/conf.h>
#include <stropts.h>

#ifndef IS_REG
#define IS_REG(x) (((x)&S_IFMT)&(S_IFREG))
#endif

/* Global variable.. Could be used for some status info, perhaps?
 * At the moment it is used when the shuffler should exit. It won't until all
 * threads are done (all data is written to the clients, or client aborted)
 */
int numthreads; 


inline int file_size(int fd)
{
  struct stat tmp;
  int res;

  if(isastream(fd)) return -2;
  if(!fstat(fd, &tmp)) return res = tmp.st_size;
  return -1;
}

/* my_write and my_read is somewhat nicer than the ones built into the system.
 * They won't give up, unless the client really abort, or the source is empty. 
 */
inline int my_write(int to, char *buf, int towrite)
{
  int res;
  while(towrite)
  {
    while((res = write(to, buf, towrite)) < 0)
    {
      switch(errno)
      {
       case EAGAIN:
       case EINTR:
	thr_yield();
	continue;

       default:
	if(errno != EPIPE)
	  perror("Shuffle: While writing");
	res = 0;
	return 0;
      }
    }
    towrite -= res;
    buf += res;
  }
  return res;
}

inline int my_read(int from, char *buf, int towrite)
{
  int res;
  while((res = read(from, buf, towrite)) < 0)
  {
    switch(errno)
    {
     case EAGAIN:
     case EINTR:
      thr_yield();
      continue;

     default:
      perror("Shuffle: While reading");
      res = 0;
      return 0;
    }
  }
  return res;
}

/* Magic defines and uggly variable passing. Don't look.
 * This won't work if a pointer is smaller than 32 bits, but I think it 
 * will work if an integer is 64, but the pointer only 32. The reverse is 
 * most definately true.
 *
 * This method will also fail utterly with fd's larger than 65535. Not very
 * likely, since you may only have 1024 on solaris 2.4. 
 *
 * The only reason for this magic is that the tread start function 
 * takes an void pointer as the only argument, and I don't really want to 
 * malloc 16 bytes for each call to this function, and it looked like to
 * much trouble to maintain a list of free blocks on my own, especially in
 * a threaded environment..
 */

#define fromfd  (((int)fromto) >> 16)
#define tofd (((int)fromto) & 65535)

void *shuffle(void *fromto)
{
  int towrite, orig;
  char *mmapped = NULL, *buffer = NULL;

  if((orig=towrite=file_size(fromfd)) == -1)
  {
    fprintf(stderr, "Shuffle: Source is not useful!\n", 0, 0);
    perror("Shuffle: file_size");
    close(tofd);
    close(fromfd);
    return;
  } 

  /* If the file is smaller than 8Kb we don't even consider mmap()
   * with friends. It is slower than read() and write(). 
   *
   * The turnaround point is really around 16Kb, but I only malloc() 8Kb,
   * so files smaller than 8Kb can be read in one read().
   */

  if(towrite > 8192)
  { 
    /* Smaller files get higher priority. This should probably be
     * a somewhat smarter routine, with more priority levels.. 
     * Unlike process priority, thread priority is higher the higher the
     * value. Probably only to confuse me :-)
     */
    if(towrite > 65535) 
      thr_setprio(1, thr_self());
    else
      thr_setprio(2, thr_self());

    if((mmapped = mmap(0, towrite, PROT_READ, MAP_SHARED, fromfd, 0))
       == MAP_FAILED)
    {
      perror("Shuffle: mmap failed");
      mmapped = 0;
    } else {
      close(fromfd);
      /* This is the actual data moving... All in one line :-) */
      my_write(tofd, mmapped, towrite);
      munmap(mmapped, orig);
    }
  }

  if(!mmapped)
  {
    buffer = malloc(8192);
    if(!buffer)
    {
      fprintf(stderr, "Shuffle: Out of memory while shuffling.\n");
      close(tofd);
      close(fromfd);
      return 0;
    }
    while(1)
    {
      towrite = my_read(fromfd, buffer, 8192);
      if(!towrite) break;
      my_write(tofd, buffer, towrite);
    }
    free(buffer);
    close(fromfd);
  }
  /* Now we are more or less done, it won't matter if the 
   * program exit() at this point, since the close and return and
   * all the other things done at thr_exit() will be done in exit()..
   */
  numthreads--;

  close(tofd);
  return 0;
}


int send_data(int from, int to)
{
  thread_t foo_thread;
  int res;
  int args;

  args = (from << 16) + to;

  if(from <= 0 || to <= 0)
  {
    fprintf(stderr, "Shuffle: Illegal fd to send_data. (%d/%d)\n", from, to);
    return 0;
  }
  
  while(res=thr_create(NULL, 0, shuffle, (void *)args, 
 		       THR_DETACHED, &foo_thread))
  {
    switch(res)
    {
     case EAGAIN:
      break;

     case ENOMEM:
      perror("Shuffle: While creating thread");
      break;

     default:
      perror("Shuffle: While creating thread");
      fprintf(stderr, "Aborting.\n");
      return 0;
    }
  }
  numthreads++;
  return 1;
}


void abort_it()
{
  int j;
  fprintf(stderr, "Shuffle: Waiting for threads to finish..");
  while(numthreads && ++j<200) 
  {
    fprintf(stderr, "(%d) ", numthreads);
    sleep(1);
  }
  if(j<200)
    fprintf(stderr, "Exiting.\n");
  else
    fprintf(stderr, "Timeout on threads, Exiting.\n");
  exit(0);
}

/* Magic, really. One could wish for a somewhat more beautiful way to
 * do this, but the ways of SYSV is not for mortals to understand.  
 */

#if 0
int receive_fd()
{
  int fd = -1;
  struct iovec  iov;
  struct msghdr msg;

  iov.iov_base         = NULL;
  iov.iov_len          = 0;
  msg.msg_iov          = &iov;
  msg.msg_iovlen       = 1;
  msg.msg_name         = NULL;
  msg.msg_accrights    = (caddr_t)&fd;
  msg.msg_accrightslen = sizeof(fd);
  
  errno = 0;
  while(recvmsg(0, &msg, 0) == -1)
  {
    switch(errno)
    {
     case EIO:  case EBADF:  case ENOTSOCK:  case EINVAL: case ESTALE:
      perror("Shuffle: Cannot receive fd from stdin");
      abort_it();

     case EWOULDBLOCK: case EINTR: case ENOMEM: case ENOSR:
      continue;

     default:
      perror("Shuffle: Unknown error while recvmsg()ing");
      abort_it();
    }
  }
  
  if(fd < 0)
  {
    fprintf(stderr, "Shuffle: Got EOF from server.\n");
    abort_it();
  }
  return fd;
}
#else
int receive_fd()
{
  struct strrecvfd tmp;
  while(ioctl(0, I_RECVFD, &tmp) == -1)
  {
    switch(errno)
    {
     case EIO:  case EBADF:  case ENOTSOCK:  case EINVAL: case ESTALE:
    case EBADMSG:
      perror("Shuffle: Cannot receive fd from stdin");
      abort_it();

     case EWOULDBLOCK: case EINTR: case ENOMEM: case ENOSR:
      continue;

    case EMFILE:
      perror("Unable to allocate another FD.\n");
      
      break;

     default:
      perror("Shuffle: Unknown error while recvmsg()ing");
      abort_it();
    }
  }
  return tmp.fd;
}
#endif

 
int main(int argc, char **argv)
{
  while(1) send_data(receive_fd(), receive_fd());
}



